! MDSBI version 4.1.7 -- Multi-Dimensional Spectral Boundary Integral Code
module io

  ! IO contains variables and routines for input/output
  ! 
  ! Modified: 25 October 2007
  
  use constants, only : pr,pin

  implicit none

  ! Global Parameters:
  ! NSTATUS = unit of status file
  ! NERROR = unit of error file

  integer(pin),save :: nstatus,nerror

  interface write_matlab
     ! WRITE_MATLAB overloads the subroutine name for writing
     ! various types of data in matlab format
     module procedure write_matlab_character
     module procedure write_matlab_logical
     module procedure write_matlab_integer
     module procedure write_matlab_integer8
     module procedure write_matlab_real
  end interface

contains


  subroutine flush_io(unit)
    ! FLUSH_IO flushes the file unit (forces data to be written now, rather
    ! than queued up--whether or not this is done automatically depends
    ! on your compiler).  Note that this is not Fortran 95 standard (but is Fortran 2003)
    ! so you may need to comment it if you get a compiler error.  So you don't
    ! have to track this down in many places, I've centralized the call to this
    ! subroutine.
    !
    ! Modified: 29 May 2007

    implicit none

    ! I/O Parameters:
    ! UNIT = file unit to be flushed

    integer(pin),intent(in) :: unit

    ! uncomment one of the following lines for immediate data output,
    ! comment the following lines for maximum speed or if your compiler complains

    call flush(unit) ! works with most compilers
    !flush(unit) ! Fortran 2003 standard

  end subroutine flush_io


  function new_io_unit() result(next)
    ! NEW_IO_UNIT returns the next available file unit
    ! 
    ! Modified: 20 February 2005

    implicit none

    ! I/O Parameters:
    ! NEXT = unit number of new file

    integer(pin) :: next

    ! Internal Parameters:
    ! MIN = minimum unit number
    ! MAX = maximum unit number
    ! LAST = unit number that was last returned

    integer(pin),parameter :: min=10,max=999
    integer(pin),save :: last=10
    logical :: open,top=.false.

    ! start at last opened unit and move up, checking if file unit
    ! is in use; if not, then return that number

    next = min-1
    if (last>0) then
       next = last+1
       inquire(unit=next,opened=open)
       if (.not.open) last = next
       return
    else
       do
          next = next+1
          inquire(unit=next,opened=open)
          if (.not.open) then
             last = next
             exit
          end if
          if (next==max) then
             if (top) call error('No available file unit','new_io_unit')
             top = .true.
             next = min-1
          end if
       end do
    end if

  end function new_io_unit


  subroutine open_status_file(status_name,error_name)
    ! OPEN_STATUS_FILE opens a file unit for the status and error files
    ! 
    ! Modified: 21 September 2007

    use mpi_routines, only : is_master,finish_mpi

    implicit none

    ! I/O Parameters:
    ! STATUS_NAME = name of status file
    ! ERROR_NAME = name of error file

    character(*),intent(in),optional :: status_name,error_name

    ! Internal Parameters:
    ! SNAME = name of status file
    ! ENAME = name of error file
    ! STAT = I/O error flag

    character(64) :: sname,ename
    integer(pin) :: stat

    if (.not.is_master()) return

    ! get unit number for files and open them

    nstatus = new_io_unit()
    nerror = new_io_unit()

    if (present(status_name)) then
       sname = status_name
    else
       sname = 'status'
    end if

    if (present(error_name)) then
       ename = error_name
    else
       ename = 'error'
    end if

    open(nstatus,file=sname,iostat=stat,status='replace')

    ! error trap opening

    if (stat/=0) then
       write(0,'(a)') "Error opening status file: '" // trim(adjustl(sname)) // "'"
       call finish_mpi
       stop
    end if

    open(nerror,file=ename,iostat=stat,status='replace')

    ! error trap opening

    if (stat/=0) then
       write(0,'(a)') "Error opening error file: '" // trim(adjustl(ename)) // "'"
       call finish_mpi
       stop
    end if

  end subroutine open_status_file


  subroutine close_status_file()
    ! CLOSE_STATUS_FILE closes file unit for status and error files
    ! 
    ! Modified: 21 September 2007
    
    use mpi_routines, only : is_master

    implicit none
    
    ! I/O Parameters:
    ! STAT = I/O error flag

    integer(pin) :: stat

    if (.not.is_master()) return

    ! close file unit for status file

    close(nstatus,iostat=stat)

    ! error trap closing

    if (stat/=0) call error('Error closing status file','close_status_file')

    ! close file unit for error file

    close(nerror,iostat=stat)

    ! error trap closing

    if (stat/=0) call error('Error closing error file','close_status_file')

  end subroutine close_status_file


  subroutine write_status(str,time,date,overwrite,file)
    ! WRITE_STATUS writes a string to the status file, with optional
    ! date and time stamps, and overwrite feature
    ! 
    ! Modified: 29 May 2007

    implicit none

    ! I/O Parameters:
    ! STR = string to be written to status file
    ! DATE = write date to file
    ! TIME = write time to file
    ! OVERWRITE = instead of writing string on a new line, 
    ! overwrite the last existing line in status file
    ! FILE = file for output ('status' or 'error')

    character(*),intent(in),optional :: str,file
    logical,intent(in),optional :: date,time,overwrite

    ! Internal Parameters:
    ! D = date
    ! T = time
    ! UNIT = file unit for output

    character(8) :: d
    character(10) :: t
    integer(pin) :: unit

    ! output unit

    if (present(file)) then
       select case(file)
       case('error')
          unit = nerror
       case default
          unit = nstatus
       end select
    else
       unit = nstatus
    end if

    ! if the last line is to be overwritten, back up in file

    if (present(overwrite)) then
       if (overwrite) backspace(unit)
    end if

    ! get date and time if needed

    if (present(date).or.present(time)) call date_and_time(d,t)

    ! write date and/or time to file

    if (present(date).and.present(time)) then
       write(unit,'(a)',advance='no') &
            t(1:2) // ':' // t(3:4) // ':' // t(5:6) // ' ' // d(1:4) // '/' // d(5:6) // '/' // d(7:8) // ' '
    else
       if (present(time)) write(unit,'(a)',advance='no') t(1:2) // ':' // t(3:4) // ':' // t(5:6) // ' '
       if (present(date)) write(unit,'(a)',advance='no') d(1:4) // '/' // d(5:6) // '/' // d(7:8) // ' '
    end if

    ! write string to file

    if (present(str)) write(unit,'(a)') trim(adjustl(str))

    ! flush file unit

    call flush_io(unit)

  end subroutine write_status

  
  subroutine error(str,routine)
    ! ERROR writes error message and terminates program
    ! 
    ! Modified: 19 October 2007

    use mpi_routines, only : my_rank
    use mpi

    implicit none

    ! I/O Parameters:
    ! STR = string to be written to error file
    ! ROUTINE =  subroutine name in which error occurred

    character(*),intent(in) :: str
    character(*),intent(in),optional :: routine
   
    ! Internal Parameters:
    ! IERR = MPI error flag
    ! PRC_STR = string holding process number

    integer(pin) :: ierr
    character(64) :: prc_str

    ! write error message and subroutine name (if given)
    
    write(prc_str,'(a,i6)') 'Message from process ',my_rank()

    call write_status(trim(adjustl(prc_str)),time=.true.,date=.true.,file='error')
    if (present(routine)) &
         call write_status('Error in subroutine: ' // trim(adjustl(routine)),time=.true.,date=.true.,file='error')
    call write_status(trim(adjustl(str)) ,time=.true.,date=.true.,file='error')
    call write_status('Terminating program, see error file for messages',time=.true.,date=.true.)

    ! terminate program

    call MPI_Abort(MPI_COMM_WORLD,0,ierr)

  end subroutine error


  subroutine warning(str,routine)
    ! WARNING writes error message, but does not terminate program
    ! 
    ! Modified: 19 October 2007

    implicit none

    ! I/O Parameters:
    ! STR = string to be written to error file
    ! ROUTINE =  subroutine name in which error occurred

    character(*),intent(in) :: str
    character(*),intent(in),optional :: routine
   
    ! write error message and subroutine name (if given)
    
    if (present(routine)) &
         call write_status('Warning in subroutine: ' // trim(adjustl(routine)),time=.true.,date=.true.,file='error')
    call write_status(trim(adjustl(str)),time=.true.,date=.true.,file='error')

  end subroutine warning


  subroutine open_array_data(name,unit,ascii)
    ! OPEN_ARRAY_DATA opens file containing array data
    ! 
    ! Modified: 25 October 2007

    implicit none

    ! I/O Parameters:
    ! NAME = file name
    ! UNIT = file unit
    ! ASCII = input file format
    !      T = ASCII text file
    !      F = binary file

    character(*),intent(in) :: name
    integer(pin),intent(out) :: unit
    logical,intent(in) :: ascii

    ! Internal Parameters:
    ! STAT = I/O error flag
    ! FORM = input file format

    integer(pin) :: stat
    character(64) :: form

    unit = new_io_unit()

    if (ascii) then
       form = 'formatted'
    else
       form = 'unformatted'
    end if

    ! open data file, call error routine if needed
    
    open(unit,file=name,form=form,iostat=stat,status='old')
    if (stat/=0) call error("Error opening file '" // trim(adjustl(name)) // "'",'open_array_data')
       
  end subroutine open_array_data


  subroutine close_array_data(unit)
    ! CLOSE_ARRAY_DATA closes file containing array data
    ! 
    ! Modified: 15 March 2005

    implicit none

    ! I/O Parameters:
    integer(pin),intent(in) :: unit

    ! Internal Parameters:
    ! STAT = I/O error flag

    integer(pin) :: stat

    close(unit,iostat=stat)

    ! error trap closing

    if (stat/=0) call error('Error closing array data file','close_array_data')

  end subroutine close_array_data


  subroutine read_array_data(unit,data,ascii,arrname)
    ! READ_ARRAY_DATA reads array data from file
    ! 
    ! Modified: 26 February 2006

    implicit none

    ! I/O Parameters:
    ! UNIT = file unit
    ! DATA = data array
    ! ASCII = input file format
    !      T = ASCII text file
    !      F = binary file
    ! ARRNAME = name of data array

    integer(pin),intent(in) :: unit
    real(pr),dimension(:,:),intent(inout) :: data
    logical,intent(in) :: ascii
    character(*),intent(in),optional :: arrname

    ! Internal Parameters:
    ! STAT = I/O error flag
    ! NAME = file name
    ! STR = error string

    integer(pin) :: stat
    character(64) :: name
    character(128) :: str

    ! read data array

    if (ascii) then
       read(unit,*,iostat=stat) data
    else
       read(unit,iostat=stat) data
    end if

    ! call error routine if needed
    if (stat/=0) then
       inquire(unit,name=name)
       if (present(arrname)) then
          str = "Error reading array '" // trim(adjustl(arrname)) // &
               "' from file '" // trim(adjustl(name)) // "'"
       else
          str = "Error reading array from file '" // trim(adjustl(name)) // "'"
       end if
       call error(str,'read_array_data')
    end if
    
  end subroutine read_array_data


  subroutine write_matlab_character(nunit,name,str,struct)
    ! WRITE_MATLAB_CHARACTER writes character string to matlab file
    ! 
    ! Modified: 20 February 2005

    implicit none

    ! I/O Parameters:
    ! NUNIT = unit number of output file
    ! NAME = name of variable
    ! STR = value of variable
    ! STRUCT = name of structure in which to write variable

    integer(pin),intent(in) :: nunit
    character(*),intent(in) :: name,str
    character(*),intent(in),optional :: struct

    ! Internal Parameters:
    ! STAT = I/O error flag

    integer(pin) :: stat

    ! write character string to file

    if (present(struct)) then
       write(nunit,'(a)',iostat=stat) &
            trim(adjustl(struct)) // '.' // trim(adjustl(name)) // " = '" // trim(adjustl(str)) // "';"
    else
       write(nunit,'(a)',iostat=stat) &
            trim(adjustl(name)) // " = '" // trim(adjustl(str)) // "';"
    end if

    if (stat/=0) call error('Error writing ' // trim(adjustl(name)) // ' to matlab file','write_matlab_character')

  end subroutine write_matlab_character


  subroutine write_matlab_integer(nunit,name,data,struct)
    ! WRITE_MATLAB_INTEGER writes integer to matlab file
    ! 
    ! Modified: 29 August 2006

    implicit none

    ! I/O Parameters:
    ! NUNIT = unit number of output file
    ! NAME = name of variable
    ! DATA = value of variable
    ! STRUCT = name of structure in which to write variable

    integer(pin),intent(in) :: nunit,data
    character(*),intent(in) :: name
    character(*),intent(in),optional :: struct

    ! Internal Parameters:
    ! STAT = I/O error flag

    integer(pin) :: stat
    
    ! write integer to file

    if (present(struct)) then
       write(nunit,'(a,i12,a)',iostat=stat) &
            trim(adjustl(struct)) // '.' // trim(adjustl(name)) // ' = ',data,';'
    else
       write(nunit,'(a,i12,a)',iostat=stat) &
            trim(adjustl(name)) // ' = ',data,';'
    end if
    
    if (stat/=0) call error('Error writing ' // trim(adjustl(name)) // ' to matlab file','write_matlab_character')

  end subroutine write_matlab_integer


  subroutine write_matlab_integer8(nunit,name,data,struct)
    ! WRITE_MATLAB_INTEGER8 writes 8-byte integer to matlab file
    ! 
    ! Modified: 29 August 2006

    use constants, only : int8
    implicit none

    ! I/O Parameters:
    ! NUNIT = unit number of output file
    ! NAME = name of variable
    ! DATA = value of variable
    ! STRUCT = name of structure in which to write variable

    integer(pin),intent(in) :: nunit
    integer(int8),intent(in) :: data
    character(*),intent(in) :: name
    character(*),intent(in),optional :: struct

    ! Internal Parameters:
    ! STAT = I/O error flag

    integer(pin) :: stat
    
    ! write integer to file

    if (present(struct)) then
       write(nunit,'(a,i12,a)',iostat=stat) &
            trim(adjustl(struct)) // '.' // trim(adjustl(name)) // ' = ',data,';'
    else
       write(nunit,'(a,i12,a)',iostat=stat) &
            trim(adjustl(name)) // ' = ',data,';'
    end if
    
    if (stat/=0) call error('Error writing ' // trim(adjustl(name)) // ' to matlab file','write_matlab_character')

  end subroutine write_matlab_integer8


  subroutine write_matlab_real(nunit,name,data,struct)
    ! WRITE_MATLAB_REAL writes real to matlab file
    ! 
    ! Modified: 20 February 2005

    implicit none

    ! I/O Parameters:
    ! NUNIT = unit number of output file
    ! NAME = name of variable
    ! DATA = value of variable
    ! STRUCT = name of structure in which to write variable

    integer(pin),intent(in) :: nunit
    character(*),intent(in) :: name
    real(pr),intent(in) :: data
    character(*),intent(in),optional :: struct

    ! Internal Parameters:
    ! STAT = I/O error flag

    integer(pin) :: stat

    ! write real to file

    if (present(struct)) then
       write(nunit,'(a,e14.6,a)',iostat=stat) &
            trim(adjustl(struct)) // '.' // trim(adjustl(name)) // ' = ',data,';'
    else
       write(nunit,'(a,e14.6,a)',iostat=stat) &
            trim(adjustl(name)) // ' = ',data,';'
    end if

    if (stat/=0) call error('Error writing ' // trim(adjustl(name)) // ' to matlab file','write_matlab_character')

  end subroutine write_matlab_real


  subroutine write_matlab_logical(nunit,name,data,struct)
    ! WRITE_MATLAB_LOGICAL writes logical to matlab file
    ! 
    ! Modified: 16 June 2007

    implicit none

    ! I/O Parameters:
    ! NUNIT = unit number of output file
    ! NAME = name of variable
    ! DATA = value of variable
    ! STRUCT = name of structure in which to write variable

    integer(pin),intent(in) :: nunit
    character(*),intent(in) :: name
    logical,intent(in) :: data
    character(*),intent(in),optional :: struct

    ! Internal Parameters:
    ! STR = string to hold 'T' or 'F'

    character(1) :: str

    ! assign logical value to string

    if (data) then
       str = 'T'
    else
       str = 'F'
    end if

    ! write string to matlab file

    if (present(struct)) then
       call write_matlab_character(nunit,name,str,struct)
    else
       call write_matlab_character(nunit,name,str)
    end if

  end subroutine write_matlab_logical


  subroutine save_data(nunit,name,data,nx,ny)
    ! SAVE_DATA writes a 2D array to a file in ascii format
    ! 
    ! Modified: 3 November 2005

    implicit none
    
    ! I/O Parameters:
    ! NUNIT = unit number of output file
    ! NAME = filename (to which .dat will be appended)
    ! DATA = data array
    ! NX = number of points in x direction
    ! NY = number of points in y direction

    integer,intent(in) :: nunit,nx,ny
    character(*),intent(in) :: name
    real(kind=pr),dimension(nx,ny),intent(in) :: data

    ! Internal Parameters:
    ! FILENAME = filename
    ! FMT = ascii format
    ! I = index in x direction
    ! J = index in y direction

    character(64) :: filename,fmt
    integer :: j

    filename = trim(adjustl(name)) // '.dat'

    open(nunit,file=filename,form='formatted')
    
    write(fmt,'(i12)') nx
    fmt = '(' // trim(adjustl(fmt)) // 'e16.7)'
    
    do j=1,ny
       write(nunit,trim(fmt)) data(:,j)
    end do
    
    close(nunit)
    
  end subroutine save_data


end module io
