! MDSBI version 4.1.7 -- Multi-Dimensional Spectral Boundary Integral Code
module thermpres

  ! THERMPRES contains variables and routines related to the
  ! thermal pressurization friction law
  ! 
  ! Modified: 19 September 2007

  use constants, only : pr,pin

  implicit none


  ! THERMPRES_TYPE is a derived type containing variables related to 
  ! thermal pressurization friction law
  !
  ! Parameters:
  ! K = thermal conductivity
  ! ALPHATH = thermal diffusivity
  ! ALPHAHY = hydraulic diffusivity
  ! LAMBDA = pore pressure rise per unit temprature rise
  ! SHAPE = shape of strain distribution
  ! W = width of deformation zone
  ! DTDZ = initial gradient in temperature, used to initialize temperture profile
  !      and used as the outer boundary condition on the diffusion equation

  type thermpres_type
     real(pr) :: K,alphath,alphahy,Lambda,dTdz,W
     character(64) :: shape
  end type thermpres_type


contains


  subroutine read_thermpres(ninput,tp)
    ! READ_THERMPRES reads in thermpres variables from file
    ! 
    ! Modified: 30 May 2007

    use constants, only : zero,half
    use io, only : error
    use mpi_routines, only : master,is_master,MPI_REAL_PR
    use mpi

    implicit none
    
    ! I/O Parameters:
    ! NINPUT = unit number for *.in input file
    ! TP = thermal pressurization variables

    integer(pin),intent(in) :: ninput
    type(thermpres_type),intent(out) :: tp

    ! Internal Parameters:
    ! STAT = I/O error flag
    ! K = thermal conductivity
    ! ALPHATH = theamal diffusivity
    ! ALPHAHY = hydraulic diffusivity
    ! LAMBDA = pore pressure rise per unit temerature rise
    ! DTDZ = initial gradient in temperature
    ! SHAPE = shape of strain distribution
    ! W = width of deformation zone
    ! IERR = MPI error flag

    integer(pin) :: stat,ierr
    real(pr) :: K,alphath,alphahy,Lambda,dTdz,W
    character(64) :: shape

    ! make namelist of user input variables

    namelist /thermpres_list/ K,alphath,alphahy,Lambda,dTdz,shape,W

    if (is_master()) then

       ! assign default values
       
       K = 2.07d-3
       alphath = 0.675_pr
       alphahy = 0.61_pr
       Lambda = 2.8_pr
       dTdz = zero
       shape = 'plane'
       W = half
       
       ! read namelist from input file, call error routine if needed
       
       rewind(ninput)
       read(ninput,nml=thermpres_list,iostat=stat)
       if (stat>0) call error("Error reading namelist 'thermpres_list' in .in file",'read_thermpres')
       
       ! assign input variables to components of derived type
       
       tp%K = K
       tp%alphath = alphath
       tp%alphahy = alphahy
       tp%Lambda = Lambda
       tp%dTdz = dTdz
       tp%shape = shape
       tp%W = W

    end if

    ! distribute variables

    call MPI_Bcast(tp%K      ,1,MPI_REAL_PR,master,MPI_COMM_WORLD,ierr)
    call MPI_Bcast(tp%alphath,1,MPI_REAL_PR,master,MPI_COMM_WORLD,ierr)
    call MPI_Bcast(tp%alphahy,1,MPI_REAL_PR,master,MPI_COMM_WORLD,ierr)
    call MPI_Bcast(tp%Lambda ,1,MPI_REAL_PR,master,MPI_COMM_WORLD,ierr)
    call MPI_Bcast(tp%dTdz   ,1,MPI_REAL_PR,master,MPI_COMM_WORLD,ierr)
    call MPI_Bcast(tp%W      ,1,MPI_REAL_PR,master,MPI_COMM_WORLD,ierr)
    call MPI_Bcast(tp%shape,64,MPI_CHARACTER,master,MPI_COMM_WORLD,ierr)

  end subroutine read_thermpres


  subroutine init_thermpres(necho,mdl,flt,tp)
    ! INIT_THERMPRES initializes thermpres variables
    ! 
    ! Modified: 21 July 2007

    use constants, only : zero
    use io, only : write_matlab
    use model, only : model_type
    use fault, only : fault_type
    use mpi_routines, only : is_master

    implicit none
    
    ! I/O Parameters:
    ! NECHO = unit number for output file
    ! MDL = model variables
    ! FLT = fault variables
    ! TP = thermal pressurization variables

    integer(pin),intent(in) :: necho
    type(model_type),intent(in) :: mdl
    type(thermpres_type),intent(in) :: tp
    type(fault_type),intent(inout) :: flt

    ! Internal Parameters:
    ! K = index in z direction
    ! TO = initial profile of temperature

    integer(pin) :: k
    real(pr),dimension(:),allocatable :: To

    ! allocate and initialize diffusion grids

    allocate(flt%T( mdl%nz,mdl%mx:mdl%px,mdl%ny,mdl%ns))
    allocate(flt%dT(mdl%nz,mdl%mx:mdl%px,mdl%ny,mdl%ns))
    allocate(flt%p( mdl%nz,mdl%mx:mdl%px,mdl%ny,mdl%ns))
    allocate(flt%dp(mdl%nz,mdl%mx:mdl%px,mdl%ny,mdl%ns))
    
    ! initialize temperature with linear profile and pressure as zero

    flt%T = zero
    flt%dT = zero
    flt%p = zero
    flt%dp = zero

    ! add linear profile

    allocate(To(mdl%nz))

    To = tp%dTdz*mdl%z

    forall (k=1:mdl%nz)
       flt%T(k,:,:,:) = flt%T0+To(k)
    end forall

    deallocate(To)

    ! output variable values into matlab file

    if (is_master()) then

       call write_matlab(necho,'K',tp%K,'tp')
       call write_matlab(necho,'alphath',tp%alphath,'tp')
       call write_matlab(necho,'alphahy',tp%alphahy,'tp')
       call write_matlab(necho,'Lambda',tp%Lambda,'tp')
       call write_matlab(necho,'dTdz',tp%dTdz,'tp')
       call write_matlab(necho,'shape',tp%shape,'tp')
       call write_matlab(necho,'W',tp%W,'tp')

    end if

  end subroutine init_thermpres


  subroutine destroy_thermpres(flt)
    ! DESTROY_THERMPRES destroys derived type tp
    ! 
    ! Modified: 19 September 2007

    use fault, only : fault_type

    implicit none

    ! I/O Parameters:
    ! FLT = fault variables

    type(fault_type),intent(inout) :: flt

    if (allocated(flt%T)) deallocate(flt%T)
    if (allocated(flt%dT)) deallocate(flt%dT)
    if (allocated(flt%p)) deallocate(flt%p)
    if (allocated(flt%dp)) deallocate(flt%dp)

  end subroutine destroy_thermpres


  subroutine thermpres_rate(mdl,flt,tp,stage)
    ! THERMPRES_RATE solves the 1D diffusion equations for thermal pressurization
    ! using explicit finite differences
    ! 
    ! Modified: 21 July 2007

    use model, only : model_type
    use fault, only : fault_type

    implicit none
    
    ! I/O Parameters:
    ! MDL = model variables
    ! FLT = fault variables
    ! TP = thermal pressurization variables
    ! STAGE = integration step

    type(model_type),intent(in) :: mdl
    type(fault_type),intent(inout) :: flt
    type(thermpres_type),intent(in) :: tp
    integer(pin),intent(in) :: stage

    ! Internal Parameters:
    ! I = index in x direction
    ! J = index in y direction
    ! Q = shear heating rate

    integer(pin) :: i,j
    real(pr),dimension(:,:),allocatable :: Q

    ! return if not including thermal pressurization model
    
    if (.not.flt%thermpres) return
       
    ! return if process holds no x data 

    if (.not.mdl%holds_x) return

    ! calculate shear heat

    allocate(Q(mdl%mx:mdl%px,mdl%ny))
    
    !$OMP PARALLEL DEFAULT(SHARED) PRIVATE(i,j)

    !$OMP DO

    do j = 1,mdl%ny
       do i = mdl%mx,mdl%px
          
          Q(i,j) = shear_heat(flt,i,j,stage)
          
       end do
    end do
    
    !$OMP END DO

    !$OMP END PARALLEL
    
    ! loop over the fault plane and solve diffusion equations
    
    !$OMP PARALLEL DEFAULT(SHARED) PRIVATE(i,j)

    !$OMP DO

    do j = 1,mdl%ny
       do i = mdl%mx,mdl%px
                 
          call thermpres_rate_point(mdl,flt%dT(:,i,j,stage),flt%dp(:,i,j,stage), &
               flt%T(:,i,j,stage),flt%p(:,i,j,stage),Q(i,j),tp)
          
       end do
    end do

    !$OMP END DO

    !$OMP END PARALLEL
    
    ! deallocate memory

    deallocate(Q)

  end subroutine thermpres_rate


  subroutine thermpres_rate_point(mdl,dT,dp,T,p,Q,tp)
    ! THERMPRES_RATE_POINT solves the 1D diffusion equations for thermal pressurization
    ! using explicit finite differences at one point on fault
    ! 
    ! Modified: 2 March 2007

    use model, only : model_type
    use constants, only : zero,half,one,two,twopi
    use asperity, only : asperity_type,ampl
    use io, only : error

    implicit none

    ! I/O Parameters:
    ! MDL = model variables
    ! DT = rate of change of temperature
    ! DP = rate of change of pore pressure
    ! T = temperature
    ! P = pore pressure
    ! Q = shear heating rate (slip velocity times shear stress)
    ! TP = thermal pressurization variables

    type(model_type),intent(in) :: mdl
    real(pr),dimension(:),intent(out) :: dT,dp
    real(pr),dimension(:),intent(in) :: T,p
    real(pr),intent(in) :: Q
    type(thermpres_type),intent(in) :: tp

    ! Internal Parameters:
    ! M = length of T,p
    ! K = index in z direction
    ! SN = effective normal stress, positive in compression
    ! RT = alpha_th/dz**2
    ! RP = alpht_hy/dz**2
    ! TP1 = T(i+1), used in finite difference equations
    ! TM1 = T(i-1), used in finite difference equations
    ! PP1 = p(i+1), used in finite difference equations
    ! PM1 = p(i-1), used in finite difference equations
    ! DTDZM = derivative of T with respect to z at left boundary
    ! DTDZP = derivative of T with respect to z at right boundary
    ! DPDZM = derivative of p with respect to z at left boundary
    ! DPDZP = derivative of p with respect to z at right boundary
    ! G = nondimensional shear strain rate
    ! A = normalization for strain rate distribution
    ! Z = distance normal to fault
    ! ASP = asperity, used to evaluate strain rate distribution
    ! C2 = coefficients of second-order approximation of Laplacian

    integer(pin) :: M,k
    real(pr) :: dTdzm,dTdzp,dpdzm,dpdzp,A,rt,rp
    real(pr),dimension(size(T)) :: Tp1,Tm1,pp1,pm1
    real(pr),dimension(:),allocatable :: G
    type(asperity_type),pointer :: asp=>null()

    M = size(T)

    ! temperature gradient at fault (no heat flux by symmetry unless 
    ! shear heating enters as boundary condition--only for slip-on-plane model)

    if (tp%shape=='plane') then
       dTdzm = -half*Q/tp%K
    else
       dTdzm = zero
    end if

    ! temperature gradient at right boundary

    dTdzp = tp%dTdz

    ! no fluid flow across fault (by symmetry) or out of modeled region

    dpdzm = zero
    dpdzp = zero

    ! parameter combinations

    rt = tp%alphath/mdl%dz**2
    rp = tp%alphahy/mdl%dz**2

    ! shift T and p for use in difference formulas

    Tp1 = eoshift(T, 1,T(M-1)+two*mdl%dz*dTdzp)
    Tm1 = eoshift(T,-1,T(2  )-two*mdl%dz*dTdzm)

    pp1 = eoshift(p, 1,p(M-1)+two*mdl%dz*dpdzp)
    pm1 = eoshift(p,-1,p(2  )-two*mdl%dz*dpdzm)

    ! temperature rate

    dT = rt*(Tp1-two*T+Tm1)

    ! add volumetric heat source if needed

    if (tp%shape/='plane') then

       allocate(G(M))
       allocate(asp)

       asp%type = tp%shape
       asp%inside = .true.
       asp%x0 = zero

       ! calculate nondimensionalized strain rate, G=W*(strain rate)/V,
       ! note that integral over z of G(z) must equal W, in order that
       ! integral over z of strain rate equals V

       ! calculate coefficient to properly normalize strain rate distribution

       select case(tp%shape)
       case default
          call error('Invalid shear zone shape','thermpres_rate')
       case('box')
          ! W is half-width for consistency with 'box' in asperity routines
          A = half
          asp%length_x = tp%W
          do k = 1,M
             G(k) = A*ampl(mdl%z(k),zero,mdl%dz,one,asp,dim=2)
          end do
       case('gaussian')
          ! W is standard deviation of gaussian (see asperity.f90 for exact formula)
          A = one/sqrt(twopi)
          asp%length_x = tp%W
          do k = 1,M
             G(k) = A*ampl(mdl%z(k),zero,mdl%dz,one,asp,dim=2)
          end do
       end select

       dT = dT+tp%alphath/tp%K*Q/tp%W*G
       
       deallocate(G)
       deallocate(asp)
       asp => null()
       
    end if

    ! pressure rate

    dp = rp*(pp1-two*p+pm1)+tp%Lambda*dT

  end subroutine thermpres_rate_point


  function shear_heat(flt,i,j,stage) result(Q)
    ! SHEAR_HEAT calculates shear heating rate
    ! 
    ! Modified: 6 November 2006

    use fault, only : fault_type

    implicit none

    ! I/O Parameters:
    ! FLT = fault variables   
    ! I = index in x direction
    ! J = index in y direction
    ! STAGE = integration sub-step

    type(fault_type),intent(inout) :: flt
    integer(pin),intent(in) :: i,j,stage
    real(pr) :: Q

    ! Internal Parameters:
    ! S = amplitude of shear stress

    real(pr) :: S

    S = sqrt(flt%sx(i,j,stage)**2+flt%sy(i,j,stage)**2)

    Q = s*flt%V(i,j,stage)

  end function shear_heat


end module thermpres

