% This file contains examples of using matlab commands to
% analyze data.  It will show you how to exploit the output
% opportunities from MDSBI.

% Set "datadir" to the directory where your data is stored.  If
% you don't do this, it will use the default ../data (which is most
% likely what you will want).

datadir = '~/esag/mdsbi/';

% Set "endian" to the endian of the system on which you ran the
% code, which might be different from the one on which you are
% running matlab to analyze the data.  If you used the same machine
% and don't know the endian, set this to 'n' (for 'native').

endian = 'l';

% Now we will analyze a few specific problems.

% Store the name of the problem in "name". You can try this for any of the
% following problems: slipweak, rs_ageing, rs_slip, rs_flash.

%name = 'slipweak';
name = 'rs_ageing';
%name = 'rs_slip';
%name = 'rs_flash';

% Call initialization function "init.m". This loads information about the run 
% in the structure A. This only needs to be run once per problem, unless you have
% rerun the program with different parameters. You do not have to call init.m
% again if you you are examining the data while the program is running, and
% further time steps have occurred. You can use any variable name, not just "A".

A = init(name,datadir,endian);

% To use the default (native) endian, you can also call this as
% A = init(name,datadir);

% Now we will load the data using the function "loadmesh.m".  It loads a 
% particular field output into a data structure.  To see which field outputs are
% available for loading, type "A":

A

% This displays all of matlab variables. You see several
% structures containing fields that were output (flt_Ux, flt_Vx, x12_sx, etc.).
% We will step through a few of these.

% First we will look at the slip velocity at the hypocenter (a
% single point) as a function of time, which is named "x00_Vx".
% Let's see what is stored in this structure:

A.x00_Vx

% We see a number of components. There's "name", which is the same
% as the name of the structure, but stored as a string.  "field" is
% the field that was output. There are the field dimensions, "nx",
% "ny", "nz". There are the minimum and maximum values of the
% position coordinates ("xmin", "xmax", etc.). The program is
% already set up to use the proper coordinates, so you never have to
% figure out how to translate an index to a coordinate value.

% Now load the data:

Vxh = loadmesh(A,'x00_Vx');

% The name on the left is the new structure (if you want to rename
% it, you can).  Let's examine its contents:

Vxh

% There are a few additions to what we saw previously: "x", "y",
% "z", "t", and "Vx". These new variables are the position
% and time coordinates, and an array containing the field values.

% Plot slip velocity in the x direction versus time:

plot(Vxh.t,Vxh.Vx)
xlabel('Time (s)')
ylabel('Slip velocity (m/s)')

% This gives the slip rate at the hypocenter. You will probably see
% oscillations, which are a signature of a poorly resolved solution.
% You can improve this by running at a higher resolution. This is easily
% accomplished by setting refine=2d0 or similar in the input file.

% You can also load in only some of the time steps. There are two options to
% control this. This loads in every ten steps of the same time history:

Vxh10 = loadmesh(A,'x00_Vx',10);

% Here's a plot that shows this.

plot(Vxh.t,Vxh.Vx,'b-o',Vxh10.t,Vxh10.Vx,'r.')
xlabel('Time (s)')
ylabel('Slip velocity (m/s)')

% You can also specify a time window, say between 1 and 3 s. 
% (You must also specify the stride in time, which we set to every
% three steps here):

Vxh3w = loadmesh(A,'x00_Vx',3,[1 3]);

plot(Vxh.t,Vxh.Vx,'b-o',Vxh10.t,Vxh10.Vx,'r.',Vxh3w.t,Vxh3w.Vx,'k*')
xlabel('Time (s)')
ylabel('Slip velocity (m/s)')
xlim([0 4])

% Now look at flt_Vx, which is the slip velocity on the fault. Load it:

Vxf = loadmesh(A,'flt_Vx');

% This field has two spatial dimensions.  You can see this by
% examining its contents:

Vxf

% We can make a space-time plot:

pcolor(Vxf.x,Vxf.t,Vxf.Vx),shading flat,colorbar
xlabel('Distance (km)')
ylabel('Time (s)')
title('Slip Velocity (m/s)')

% You can make similar plots for stress:

sxf = loadmesh(A,'flt_sx');
for n=1:sxf.nt
  plot(sxf.x,sxf.sx(n,:))
  xlabel('Distance (km)')
  ylabel('Shear stress (MPa)')
  axis([-40 40 50 90]) % for slipweak, rs_ageing, rs_slip
  %axis([-40 40 20 60]) % for rs_flash
  pause(0.1)
end

% You can also examine the rupture front arrival time.

frt = loadmesh(A,'frt');

% This plots rupture front arrival time (a value of 0 typically means that the
% point did not rupture).

plot(frt.x,frt.t)
xlabel('Distance (km)')
ylabel('Rupture Time (s)')

% You can perform a similar analysis of the data for other 2D problems.

% For 3D problems, you can make other plots. For example, let's try tpv3.

name = 'tpv3';
A = init(name,datadir,endian);

% Now let's loop over each time step, and plot snapshots of slip
% velocity on the fault. The way matlab's pcolor command works is
% to use the coordinate (x,y) as the left and bottom corner of 
% the colored rectangle. To properly center the rectangle, offset it by
% half of a grid spacing. You can see the influence of the
% high stress asperity on the right, and the barrier of high
% static friction on the left. Heterogeneities like these are easy to add
% in the input file. More complicated shapes require reading in data files,
% which is illustrated for tpv101.

Vx = loadmesh(A,'flt_Vx');

for n=1:Vx.nt
  pcolor(Vx.x-Vx.dx/2,Vx.y-Vx.dy/2,squeeze(Vx.Vx(n,:,:))'),shading flat,colorbar
  axis image
  axis([-20 20 -20 5])
  xlabel('Distance along strike (km)')
  ylabel('Distance down dip (km)')
  title('Slip Velocity (m/s)')
  pause(0.1)
end

% Here is the same for shear stress.

sx = loadmesh(A,'flt_sx');

for n=1:sx.nt
  pcolor(sx.x,sx.y,squeeze(sx.sx(n,:,:))'),shading flat,colorbar
  axis image
  axis([-20 20 -20 5])
  xlabel('Distance along strike (km)')
  ylabel('Distance down dip (km)')
  title('Shear Stress (MPa)')
  pause(0.1)
end

% You can also examine the rupture front arrival time.

frt = loadmesh(A,'frt');

% This plots rupture front contours every 0.5 s.

contour(frt.x,frt.y,squeeze(frt.t)',[0:0.5:12],'b')
xlabel('Distance along strike (km)')
ylabel('Distance down dip (km)')
title('Rupture Front Arrival Time (every 0.5 s)')
  
% For tpv101, there is a complicated pattern of initial conditions
% and friction law parameters, which was input to the code using a 
% set of data files. These were generated using the Matlab script
% input_tpv101.m. Take a look at that file to see how this feature works.